"""
gref4hsi.

A python library for georeferencing and orthorectifying .
"""

__version__ = "0.0.0a"
__author__ = 'Håvard Snefjellå Løvås'
__credits__ = 'Norwegian University of Science and Technology'